import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
import random
import string
import re
import collections
random.seed(0)


def prepare_sequence(lst, idx_mapping):
    """ 
    Map individual items from `sent` using `idx_mapping`
    Return value is the same length as `sent`
    Usage: 
        >> prepare_sequence(['a', 'b', 'c'], {'a':0, 'b':1, 'c':2})
        [0, 1, 2]
    """
    idxs = []
    for item in lst:
        if item not in idx_mapping:
            assert '<unk>' in idx_mapping or 0 in idx_mapping, "cannot map unknown token:" + item
            if '<unk>' in idx_mapping:
                idxs.append(idx_mapping['<unk>'])
            else:
                idxs.append(idx_mapping[0])
        else:
            idxs.append(idx_mapping[item])
    try: 
        return torch.tensor(idxs, dtype=torch.long)
    except: 
        return idxs


class LSTMEncoder(nn.Module):
    def __init__(self, embedding_dim, hidden_dim, char_embedding_dim,
                 char_hidden_dim, char_size, vocab_size, lstm_layers=1, 
                 bidirectional=False, dropout=0):
        super(LSTMEncoder, self).__init__()
        #############################################################################
        # TODO: Define and initialize anything needed for the forward pass.
        # You are required to create a model with:
        # embedding layer: that maps words to the embedding space
        # char embedding layer: maps chars to embedding space 
        # an char level LSTM: that finds the character level embedding for a word
        # an LSTM layer: that takes the combined embeddings as input and outputs hidden states
        # Remember, this needs to be done for both context and query (our input)
        # (DO NOT apply bidirectionality to character LSTMs)
        #############################################################################
        # Encoder

        # Remember: bidirectional makes the output hidden_dim * 2

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################

    def forward(self, context, context_chars, query, query_chars):
        lstm_context_vectors = None # for each word
        lstm_query_vectors = None # for each word 
        #############################################################################
        # TODO: Implement the forward pass.
        # Given a tokenized index-mapped sentence and a character sequence as the arguments,
        # `context` and `query` are word sequences at are index mapped
        # `context_chars` and `query_chars` are char sequence of each word that are 
        #   index mapped
        # The output is very similar to HW4. 
        # Return values:
        #   `lstm_query_vectors` : Txd or Nx(d*2) if bidirectional LSTM used
        #       T is # of words in context, and d is size of hidden states
        #   `lstm_query_vectors` : Jxd or Jx(d*2) if bidirectional LSTM used
        #       J is # of words in query, and d is size of hidden states
        # 
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################
        return lstm_context_vectors, lstm_query_vectors
        

class AttentionFlow(nn.Module):
    def __init__(self, w_dim):
        """
        w_dim : is the same as 6d in the paper. Should be 6*hidden_dim if bidirectionality is True
        """
        #############################################################################
        # TODO: Define and initialize anything needed for the forward pass.
        # For this part, we need a linear layer to compute the similarity matrix (no bias according to the paper)
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################

    def forward(self, context, query):
        G = None
        #############################################################################
        # T : number of tokens in context
        # J : number of tokens in query
        # d : hidden_dimensions (context and query will have d*2 hidden dimensions if using bidirectional LSTM)
        # Parameters (from the encoder):
        #     context: of size Txd or (d*2) if using bidirectional LSTM
        #     query: of size Jxd or (d*2) if using bidirectional LSTM
        # For this part, you need to compute a similarity matrix, S
        #     S : TxJ
        # then use S to build context2query and query2context attention
        #     context2query : T x (d*2)
        #     query2context : T x (d*2) 
        #         hint: query2context will be (1, d*2) but it will need to be repeated 
        #               T times so the dimension is T x (d*2)
        # Return :: G which is the query aware context vectors of size (T x d*8)
        #     G is obtained by combining `context`, `context2query` and `query2context` 
        #       as defined in the paper.
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################

        return G


class ModelingLayer(nn.Module):
    def __init__(self, input_dim, output_dim, num_layers=2, dropout=0.2, 
        bidirectional=True):
        super(ModelingLayer, self).__init__()
        #############################################################################
        # TODO: Define and initialize anything needed for the forward pass.
        # For the modeling layer, we just need to pass our query aware context vectors 
        # You need to initialize an LSTM layer here
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################
    
    def forward(self, G):
        M = None
        #############################################################################
        # TODO: Implement the forward pass.
        # G : query aware context word embeddings 
        # returns :: of size Tx(output_dim*2) (T is # words in context)
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################
        return M


class OutputLayer(nn.Module):
    def __init__(self, fc_dim, LSTM_input_size, LSTM_output_size, num_layers=1, bidirectional=True):
        super(OutputLayer, self).__init__()
        #############################################################################
        # TODO: Define and initialize anything needed for the forward pass.
        # For the OutputLayer, we need:
        #   Linear layer (to predict start idx; no bias for these linear layers according to paper) 
        #   LSTM + Linear (to predict end idx)
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################

    def forward(self, G, M):
        start, end = None, None
        #############################################################################
        # TODO: Implement the forward pass.
        # G : query aware context word embeddings
        # M : output of modeling layer
        # returns :: `start` and `end` of size (T,) 
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################
        return start, end

 
class BiDAF(nn.Module):
    def __init__(self, embedding_dim, hidden_dim, char_embedding_dim,
                 char_hidden_dim, char_size, vocab_size, bidirectional=True, phrase_LSTM_layers=1, modeling_LSTM_layers=2, dropout=0):
        super(BiDAF, self).__init__()
        #############################################################################
        # TODO: Define and initialize anything needed for the forward pass.
        # Initialize all the modules you created so far and link their inputs and 
        #       outputs properly:
        #   LSTMEncoder
        #   AttentionFlow
        #   ModelingLayer
        #   Output
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################

    def forward(self, context, context_chars, query, query_chars):
        start, end = None, None
        #############################################################################
        # TODO: Implement the forward pass.
        # Given a tokenized index-mapped sentence and a character sequence as the arguments,
        # find the corresponding scores for tags
        # returns:: `start` and `end` of size (T,) 
        #   where, T is the number of words/tokens in context
        #############################################################################

        #############################################################################
        #                             END OF YOUR CODE                              #
        #############################################################################
        return start, end


##### From Official SQUAD Evauation evaluation script version 2.0 #####

def normalize_answer(s):
    """Lower text and remove punctuation, artcles and extra whitespace"""
    def remove_articles(text):
        regex = re.compile(r'\b(a|an|the)]b', re.UNICODE)
        try:
            return re.sub(regex, ' ', text)
        except:
            return text

    def white_space_fix(text):
        try:
            return ' '.join(text.split())
        except:
            return text

    def remove_punc(text):
        exclude = set(string.punctuation)
        return ''.join(ch for ch in text if ch not in exclude)

    def lower(text):
        try:
            return text.lower()
        except:
            return text
    
    return white_space_fix(remove_articles(remove_punc(lower(s))))

def get_tokens(s):
    if not s:
        return []
    return normalize_answer(s).split()

def compute_exact(a_gold, a_pred):
    return int(normalize_answer(a_gold) == normalize_answer(a_pred))

def compute_f1(a_gold, a_pred):
    gold_toks = get_tokens(a_gold)
    pred_toks = get_tokens(a_pred)
    common = collections.Counter(gold_toks) & collections.Counter(pred_toks)
    num_same = sum(common.values())
    if len(gold_toks) == 0 or len(pred_toks) == 0:
        # If either is no-answer, then F1 is 1 if they agree, 0 otherwise
        return int(gold_toks == pred_toks)
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(pred_toks)
    recall = 1.0 * num_same / len(gold_toks)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1
